/**
 * Query information from the HighPoint controller.
 *
 * We use the direct ioctl operations defined in the hptintf.h header file.
 *
 * The direct interface is provided through the /proc/scsi/rr640l/6 filed.
 * In order to communicate with the driver, the file is opened for write and
 * a small descriptor (HPT_IOCTL_PARAM) is written into it.
 * this descriptor contains a magic word followed by some pointers to the
 * input buffer, output buffer, return size, and the lengths of the buffers.
 * On return from the write, the buffers will have been populated.
 * Yes, it really is that icky.
 */

/* Interface version I expect to talk */
#define HPT_INTERFACE_VERSION 0x02010000

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#include "hptdatatypes.h"
#include "hptintf.h"

/* Hard coded for now */
#define HPTPROC "/proc/scsi/rr640l/6"

static int hpt_fd = -1;
static char *cache_policies[] = {
    "none",
    "write-through",
    "write-back",
};
static char *capability_flags[] = {
    "dump-metadata",
    "disk-checking"
};
static char *chip_flags[] = {
    "support-ultra-133",
    "support-ultra-150",
    "master",
    "support-sata-300",
    NULL,
    "support-ultra-66",
    "support-ultra-100",
    "hpt3xx-dpll-mode"
};
static char *controller_exflags[] = {
    "CEXF_IOPModel",
    "CEXF_SDRAMSize",
    "CEXF_BatteryInstalled",
    "CEXF_BatteryStatus",
    "CEXF_BatteryVoltage",
    "CEXF_BatteryBackupTime",
    "CEXF_FirmwareVersion",
    "CEXF_SerialNumber",
    "CEXF_BatteryTemperature",
    "CEXF_Power12v",
    "CEXF_Power5v",
    "CEXF_Power3p3v",
    "CEXF_Power2p5v",
    "CEXF_Power1p8v",
    "CEXF_Core1p8v",
    "CEXF_Core1p2v",
    "CEXF_DDR1p8v",
    "CEXF_DDR1p8vRef",
    "CEXF_CPUTemperature",
    "CEXF_BoardTemperature",
    "CEXF_FanSpeed",
    "CEXF_Core1p0v",
    "CEXF_Fan2Speed",
    "CEXF_Power1p0v",
    "CEXF_Power1p5v",
    "CEXF_SASAddress"
};


/**
 * Describe a bitfield from a value, using an array of strings for each
 * bit.
 *
 * @param[in] value     Value to render
 * @param[in] labels    Bit labels to display, an array of pointers
 * @param[in] max       Highest bit label in the array
 *
 * @return static zero-terminated string pointer for the bits
 */
const char *describe_bitfield(unsigned long   value,
                              char          **labels,
                              int             max)
{
    static char buffer[1024] = "";
    int         bit;

    if (value == 0)
        return "no flags";

    strcpy(buffer, "");

    for (bit=0; (value >> bit) > 0; bit++)
    {
        if (value & (1ul<<bit))
        {
            if (buffer[0] != '\0')
            {
                strcat(buffer, ", ");
            }
            if (bit >= max ||
                labels[bit] == NULL)
            {
                sprintf( &buffer[strlen(buffer)], "bit %i", bit);
            }
            else
            {
                strcat( buffer, labels[bit] );
            }
        }
    }

    return buffer;
}

#define BITFIELD(value, labels) (describe_bitfield(value, labels, sizeof(labels)/sizeof(labels[0])))

/**
 * Perform an internal ioctl operation on the highpoint controller.
 *
 * @param[in] op        Operation to perform (HPT_IOCTL_*)
 * @param[in] in        Input data buffer
 * @param[in] inlen     Size of the input buffer
 * @param[in] out       Output data buffer
 * @param[in] outlen    Size of the output buffer
 *
 * @return number of bytes populated in output buffer (or -1 if failed)
 */
int hpt_ioctl(int op,
              void *in,  unsigned int inlen,
              void *out, unsigned int outlen)
{
    /* FIXME: I think we need to know our architecture here, but I'm just
     * using the generic interface for now, rather than the 32bit-on-64bit
     * interface.
     */
    HPT_IOCTL_PARAM param;
    HPT_UINT        returned = -1;

    param.Magic = HPT_IOCTL_MAGIC;
    param.dwIoControlCode = op;
    param.lpInBuffer = (HPT_PTR)in;
    param.nInBufferSize = inlen;
    param.lpOutBuffer = (HPT_PTR)out;
    param.nOutBufferSize = outlen;
    param.lpBytesReturned = (HPT_PTR)&returned;

    write(hpt_fd, &param, sizeof(param));

    return returned;
}


/**
 * Open communications with the Highpoint driver.
 *
 * @retval 1 if successfully opened fd
 * @retval 0 if failed to open
 */
int hpt_open(void)
{
    int fd;

    /* If already open, do not open again */
    if (hpt_fd != -1)
        return 1;

    fd = open(HPTPROC, O_RDWR);
    if (fd < 0)
        return 0;

    hpt_fd = fd;
    return 1;
}


/**
 * Close communications with the Highpoint driver.
 */
void hpt_close(void)
{
    if (hpt_fd != -1)
    {
        close(hpt_fd);
        hpt_fd = -1;
    }
}

/**
 * Version compatibility: all versions
 * Parameters:
 *  None
 * Returns:
 *  interface version. 0 when fail.
 */
HPT_U32 hpt_get_version(void)
{
    HPT_U32 version;
    if (hpt_ioctl(HPT_IOCTL_GET_VERSION,
                  NULL, 0,
                  &version, sizeof(version)) == sizeof(version))
    {
        return version;
    }
    return 0;
}

/**
 * hpt_get_controller_count
 * Version compatibility: v1.0.0.1 or later
 * Parameters:
 *  None
 * Returns:
 *  number of controllers
 */
int hpt_get_controller_count(void)
{
    int count;
    if (hpt_ioctl(HPT_IOCTL_GET_CONTROLLER_COUNT,
                  NULL, 0,
                  &count, sizeof(count)) == sizeof(count))
    {
        return count;
    }
    return -1;
}

/*
 * hpt_get_driver_capabilities
 * Version compatibility: v1.0.0.2 or later
 * Parameters:
 *  Pointer to receive a DRIVE_CAPABILITIES structure. The caller must set
 *  dwSize member to sizeof(DRIVER_CAPABILITIES). The callee must check this
 *  member to see if it's correct.
 * Returns:
 *  0 - Success
 */
int hpt_get_driver_capabilities(PDRIVER_CAPABILITIES cap)
{
    if (hpt_ioctl(HPT_IOCTL_GET_DRIVER_CAPABILITIES,
                  NULL, 0,
                  cap, sizeof(*cap)) == sizeof(*cap))
    {
        return 0;
    }
    return -1;
}
int hpt_get_driver_capabilities_v2(PDRIVER_CAPABILITIES_V2 cap)
{
    if (hpt_ioctl(HPT_IOCTL_GET_DRIVER_CAPABILITIES_V2,
                  NULL, 0,
                  cap, sizeof(*cap)) == sizeof(*cap))
    {
        return 0;
    }
    return -1;
}

/**
 * hpt_get_controller_info
 * Version compatibility: v1.0.0.1 or later (v1)
 * Version compatibility: v2.0.0.0 or later (v2,v3)
 * Parameters:
 *  id      Controller id
 *  pInfo   pointer to CONTROLLER_INFO(_V2|_V3) buffer
 * Returns:
 *  0       Success, controller info is put into (*pInfo ).
 */
int hpt_get_controller_info(int id, PCONTROLLER_INFO pInfo)
{
    if (hpt_ioctl(HPT_IOCTL_GET_CONTROLLER_INFO,
                  &id, sizeof(id),
                  pInfo, sizeof(*pInfo)) == sizeof(*pInfo))
    {
        return 0;
    }
    return -1;
}
int hpt_get_controller_info_v2(int id, PCONTROLLER_INFO_V2 pInfo)
{
    if (hpt_ioctl(HPT_IOCTL_GET_CONTROLLER_INFO_V2,
                  &id, sizeof(id),
                  pInfo, sizeof(*pInfo)) == sizeof(*pInfo))
    {
        return 0;
    }
    return -1;
}
int hpt_get_controller_info_v3(int id, PCONTROLLER_INFO_V3 pInfo)
{
    if (hpt_ioctl(HPT_IOCTL_GET_CONTROLLER_INFO_V3,
                  &id, sizeof(id),
                  pInfo, sizeof(*pInfo)) == sizeof(*pInfo))
    {
        return 0;
    }
    return -1;
}


/* hpt_get_channel_info
 * Version compatibility: v1.0.0.1 or later
 * Parameters:
 *  id      Controller id
 *  bus     bus number
 *  pInfo   pointer to CHANNEL_INFO buffer
 * Returns:
 *  0       Success, channel info is put into (*pInfo ).
 */
int hpt_get_channel_info(int id, int bus, PCHANNEL_INFO pInfo)
{
    int param[2] = {
        id,
        bus
    };
    if (hpt_ioctl(HPT_IOCTL_GET_CHANNEL_INFO,
                  &param, sizeof(param),
                  pInfo, sizeof(*pInfo)) == sizeof(*pInfo))
    {
        return 0;
    }
    return -1;
}
int hpt_get_channel_info_v2(int id, int bus, PCHANNEL_INFO_V2 pInfo)
{
    int param[2] = {
        id,
        bus
    };
    if (hpt_ioctl(HPT_IOCTL_GET_CHANNEL_INFO_V2,
                  &param, sizeof(param),
                  pInfo, sizeof(*pInfo)) == sizeof(*pInfo))
    {
        return 0;
    }
    return -1;
}


/* hpt_get_logical_devices
 * Version compatibility: v1.0.0.1 or later
 * Parameters:
 *  pIds        pointer to a DEVICEID array
 *  nMaxCount   array size
 * Returns:
 *  Number of ID returned. All logical device IDs are put into pIds array.
 *  Note: A spare disk is not a logical device.
 */
int hpt_get_logical_devices(DEVICEID * pIds, int nMaxCount)
{
    int read = hpt_ioctl(HPT_IOCTL_GET_LOGICAL_DEVICES,
                         NULL, 0,
                         pIds, sizeof(*pIds)*nMaxCount);
    if (read <= sizeof(*pIds)*nMaxCount &&
        read % sizeof(*pIds) == 0)
    {
        return read / sizeof(*pIds);
    }
    return 0;
}



/**
 * Plain test function.
 */
int main(void)
{
    int controllers = 0;

    if (!hpt_open())
    {
        printf("Failed to open highpoint driver; is it installed? are you root?\n");
        exit(1);
    }

    printf("HPT Interface Version: %08x\n", hpt_get_version());
    controllers = hpt_get_controller_count();
    printf("Controllers: %i\n", controllers);

/* Helper macros to show the results of the capabilities calls */
#define CAP_INT(x) printf("  %-32s %d\n", #x ":", capp->x);
#define CAP_RAIDSUPPORT(n, x) \
            printf("  RAID#%2i: %-23s %s : %2d members\n", \
                   n,                                     \
                   #x,                                    \
                   cap.SupportedRAIDTypes[n] ? "Supported  " : "Unsupported", \
                   cap.MaximumArrayMembers[n]);

    {
        DRIVER_CAPABILITIES cap, *capp=&cap;
        printf("Driver capabilities:\n");
        if (hpt_get_driver_capabilities(&cap))
        {
            printf("  Unavailable\n");
        }
        else
        {
            CAP_INT(MaximumControllers);
            CAP_INT(SupportCrossControllerRAID);
            CAP_INT(MinimumBlockSizeShift);
            CAP_INT(MaximumBlockSizeShift);
            CAP_INT(SupportDiskModeSetting);
            CAP_INT(SupportSparePool);
            CAP_INT(MaximumArrayNameLength);
            CAP_INT(SupportPerformanceMonitor);
            CAP_INT(SupportVariableSectorSize);
            CAP_INT(SupportHotSwap);
            CAP_INT(HighPerformanceRAID1);
            CAP_INT(RebuildProcessInDriver);
            CAP_INT(SupportDedicatedSpare);

            CAP_RAIDSUPPORT(0, AT_UNKNOWN);
            CAP_RAIDSUPPORT(1, AT_RAID0);
            CAP_RAIDSUPPORT(2, AT_RAID1);
            CAP_RAIDSUPPORT(3, AT_RAID5);
            CAP_RAIDSUPPORT(4, AT_RAID6);
            CAP_RAIDSUPPORT(5, AT_RAID3);
            CAP_RAIDSUPPORT(6, AT_RAID4);
            CAP_RAIDSUPPORT(7, AT_JBOD);
            CAP_RAIDSUPPORT(8, AT_RAID1E);
            CAP_RAIDSUPPORT(9, undefined);
            CAP_RAIDSUPPORT(10, undefined);
            CAP_RAIDSUPPORT(11, undefined);
            CAP_RAIDSUPPORT(12, undefined);
            CAP_RAIDSUPPORT(13, undefined);
            CAP_RAIDSUPPORT(14, undefined);
            CAP_RAIDSUPPORT(15, undefined);
        }
    }

/* Policies is a bitfield for the CACHE_POLICIES supported */
#define CAP2_RAIDSUPPORT(n, x) \
            printf("  RAID#%2i: %-23s %s : %2d members : Cache=%s\n", \
                   n,                                     \
                   #x,                                    \
                   capp->SupportedRAIDTypes[n] ?          \
                        "Supported  "                     \
                      : "Unsupported",                    \
                   capp->MaximumArrayMembers[n],          \
                   BITFIELD(cap2.SupportedCachePolicies[n], cache_policies));
#define CAP2_INT(x) printf("  %-32s %d\n", #x ":", cap2.x);
#define CAP2_BITS(x,y) printf("  %-32s %s\n", #x, BITFIELD(cap2.x, y));

    {
        DRIVER_CAPABILITIES_V2 cap2;
        DRIVER_CAPABILITIES   *capp=&cap2.v1;
        printf("Driver capabilities(v2):\n");
        if (hpt_get_driver_capabilities_v2(&cap2))
        {
            printf("  Unavailable\n");
        }
        else
        {
            CAP_INT(MaximumControllers);
            CAP_INT(SupportCrossControllerRAID);
            CAP_INT(MinimumBlockSizeShift);
            CAP_INT(MaximumBlockSizeShift);
            CAP_INT(SupportDiskModeSetting);
            CAP_INT(SupportSparePool);
            CAP_INT(MaximumArrayNameLength);
            CAP_INT(SupportPerformanceMonitor);
            CAP_INT(SupportVariableSectorSize);
            CAP_INT(SupportHotSwap);
            CAP_INT(HighPerformanceRAID1);
            CAP_INT(RebuildProcessInDriver);
            CAP_INT(SupportDedicatedSpare);

            CAP2_RAIDSUPPORT(0, AT_UNKNOWN);
            CAP2_RAIDSUPPORT(1, AT_RAID0);
            CAP2_RAIDSUPPORT(2, AT_RAID1);
            CAP2_RAIDSUPPORT(3, AT_RAID5);
            CAP2_RAIDSUPPORT(4, AT_RAID6);
            CAP2_RAIDSUPPORT(5, AT_RAID3);
            CAP2_RAIDSUPPORT(6, AT_RAID4);
            CAP2_RAIDSUPPORT(7, AT_JBOD);
            CAP2_RAIDSUPPORT(8, AT_RAID1E);
            CAP2_RAIDSUPPORT(9, undefined);
            CAP2_RAIDSUPPORT(10, undefined);
            CAP2_RAIDSUPPORT(11, undefined);
            CAP2_RAIDSUPPORT(12, undefined);
            CAP2_RAIDSUPPORT(13, undefined);
            CAP2_RAIDSUPPORT(14, undefined);
            CAP2_RAIDSUPPORT(15, undefined);

            /* And now the V2 caps */
            CAP2_INT(ConfigRegSize);
            CAP2_INT(SupportDiskCachePolicy);
            CAP2_BITS(Flags, capability_flags);
        }
    }

    {
        int i,j;
        for (i = 0; i<controllers; i++)
        {
            CONTROLLER_INFO_V3 controller, *contp = &controller;
            int level = 0;
            printf("ControllerInfo(#%i):\n", i);
            if (hpt_get_controller_info_v3(i, &controller) == 0)
            {
                level = 3;
            }
            else if (hpt_get_controller_info_v2(i, (CONTROLLER_INFO_V2 *)&controller) == 0)
            {
                level = 2;
            }
            else if (hpt_get_controller_info(i, (CONTROLLER_INFO *)&controller) == 0)
            {
                level = 1;
            }
            if (level == 0)
            {
                printf("  Unavailable\n");
            }
            if (level >= 1)
            {
                printf("  %-32s %i\n", "Level supported", level);
#define CONTROLLER_INT(x) printf("  %-32s %d\n", #x, contp->x);
#define CONTROLLER_HEX(x) printf("  %-32s 0x%x\n", #x, contp->x);
#define CONTROLLER_BITS(x,y) printf("  %-32s %s\n", #x, BITFIELD(contp->x, y));
#define CONTROLLER_STR(x) printf("  %-32s '%s'\n", #x, contp->x);
                CONTROLLER_INT(ChipType);
                CONTROLLER_INT(InterruptLevel);
                CONTROLLER_INT(NumBuses);
                CONTROLLER_BITS(ChipFlags, chip_flags);
                CONTROLLER_STR(szProductID);
                CONTROLLER_STR(szVendorID);
            }
            if (level >= 2)
            {
                CONTROLLER_HEX(GroupId);
                CONTROLLER_INT(pci_tree);
                CONTROLLER_INT(pci_bus);
                CONTROLLER_INT(pci_device);
                CONTROLLER_INT(pci_function);
                CONTROLLER_BITS(ExFlags, controller_exflags);
            }
            if (level >= 3)
            {
                CONTROLLER_STR(IOPModel);
                CONTROLLER_INT(SDRAMSize);
                CONTROLLER_INT(BatteryInstalled);
                CONTROLLER_INT(BatteryStatus);
                CONTROLLER_INT(BatteryVoltage);
                CONTROLLER_INT(BatteryBackupTime);
                CONTROLLER_INT(FirmwareVersion);
                CONTROLLER_STR(SerialNumber);
                CONTROLLER_INT(BatteryMBInstalled);
                CONTROLLER_INT(BatteryTemperature);
                CONTROLLER_INT(CPUTemperature);
                CONTROLLER_INT(BoardTemperature);
                CONTROLLER_INT(FanSpeed);
                CONTROLLER_INT(Power12v);
                CONTROLLER_INT(Power5v);
                CONTROLLER_INT(Power3p3v);
                CONTROLLER_INT(Power2p5v);
                CONTROLLER_INT(Power1p8v);
                CONTROLLER_INT(Core1p8v);
                CONTROLLER_INT(Core1p2v);
                CONTROLLER_INT(DDR1p8v);
                CONTROLLER_INT(DDR1p8vRef);
                CONTROLLER_INT(Core1p0v);
                CONTROLLER_INT(Fan2Speed);
                CONTROLLER_INT(Power1p0v);
                CONTROLLER_INT(Power1p5v);
                //CONTROLLER_INT(SASAddress[8]);
                //CONTROLLER_INT(reserve[48]);
            }


            /* Channel information */
            for (j=0; j < contp->NumBuses; j++)
            {
                CHANNEL_INFO_V2 cinfo = { 0 };
                int version = 0;
                int devices = 0;
                int k;

                printf("  Channel #%i:\n", j);

                if (hpt_get_channel_info_v2(i, j, &cinfo) == 0)
                {
                    version = 2;
                }
                else if (hpt_get_channel_info(i, j, (CHANNEL_INFO *)&cinfo) == 0)
                {
                    version = 1;
                }
                if (version == 0)
                {
                    printf("    Could not read devices\n");
                }
                if (version >= 2)
                    devices = 15;
                else if (version >= 1)
                    devices = 2;

                printf("    %-32s %i\n", "Version:", version);
                printf("    %-32s 0x%08x\n", "IoPort:", cinfo.IoPort);
                printf("    %-32s 0x%08x\n", "ControlPort:", cinfo.ControlPort);
                for (k=0; k<devices; k++)
                {
                    if (cinfo.Devices[k] != 0)
                        printf("    #%-2i %-28s 0x%08x\n", k, "DeviceID", cinfo.Devices[k]);
                }
            }
        }
    }

#if 0
/* Don't do this - it causes a kernel panic */
    {
        DEVICEID ids[MAX_PHYSICAL_DEVICE];
        int devices = hpt_get_logical_devices(ids, MAX_PHYSICAL_DEVICE);
        int i;
        printf("Logical devices:\n");
        printf("  %-32s %i\n", "Count:", devices);
        for (i=0; i<devices; i++)
        {
            printf("    #%-2i %-28s 0x%08x\n", i, "DeviceID", devices[i]);
        }
    }
#endif

    hpt_close();

    return 0;
}